__SYCL_PARAM_TRAITS_SPEC(event_profiling, command_submit, uint64_t, UR_PROFILING_INFO_COMMAND_SUBMIT)
__SYCL_PARAM_TRAITS_SPEC(event_profiling, command_start, uint64_t, UR_PROFILING_INFO_COMMAND_START)
__SYCL_PARAM_TRAITS_SPEC(event_profiling, command_end, uint64_t, UR_PROFILING_INFO_COMMAND_END)

