  // -*- C++ -*-
  //===----------------------------------------------------------------------===//
  // Modifications Copyright Intel Corporation 2022
  // SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
  //===----------------------------------------------------------------------===//
  // Auto-generated from boost/mp11 sources https://github.com/boostorg/mp11

#ifndef SYCL_DETAIL_BOOST_MP11_VERSION_HPP_INCLUDED
#define SYCL_DETAIL_BOOST_MP11_VERSION_HPP_INCLUDED

//  Copyright 2019 Peter Dimov
//
//  Distributed under the Boost Software License, Version 1.0.
//
//  See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt

// Same format as SYCL_DETAIL_BOOST_VERSION:
//   major * 100000 + minor * 100 + patch

#define SYCL_DETAIL_BOOST_MP11_VERSION 108500

#endif // #ifndef SYCL_DETAIL_BOOST_MP11_VERSION_HPP_INCLUDED
